
#include "ncview.includes.h"
#include "ncview.defines.h"
#include "ncview.protos.h"

extern View  *view;
extern Options   options;

static char	overlay_filename[1024], overlay_base_dir[1024];

static void 	determine_overlay_base_dir();
static int 	gen_xform( float value, int n, float *dimvals );

/*===================================================================*/

	void
do_overlay()
{
	int message;

	if( view == NULL ) {
		x_error( "You must select a variable first." );
		options_set_overlay_state( FALSE );
		return;
		}

	if( overlay_filename == NULL ) {
		message = overlay_select_filename();
		if( message == MESSAGE_CANCEL ) 
			return;
		}

	options.overlay->overlay = gen_overlay( view, overlay_filename );
	if( options.overlay->overlay != NULL ) {
		options.overlay->doit = TRUE;
		invalidate_all_saveframes();
		change_view( 0, FRAMES );
		}
	else
		in_error( "Sorry, overlay operation failed\n" );
}

/*===================================================================*/

	void
overlay_init()
{
	FILE	*f;
	char	filename[1024];
	
	/* Select the file that we will be getting overlays from, 
	 * or NULL if we can't figure out any.
	 */
	determine_overlay_base_dir();
	sprintf( filename, "%s/%s", overlay_base_dir, DEFAULT_OVERLAY_FILENAME );
	if( (f = fopen(filename, "r")) == NULL ) {
		printf( "Note: can't open default overlay file %s\n", filename );
		overlay_filename[0] = '\0';
		}
	else
		{
		strcpy( overlay_filename, filename );
		fclose( f );
		options_set_overlay_filename( overlay_filename );
		}
}

/*===================================================================*/
	
	static void
determine_overlay_base_dir()
{
	char	*dir;

	dir = (char *)getenv( "NCVIEWBASE" );
	if( dir == NULL ) {
#ifdef NCVIEW_LIB_DIR
		strcpy( overlay_base_dir, NCVIEW_LIB_DIR );
#else
		strcpy( overlay_base_dir, "." );
#endif
		}
	else
		strcpy( overlay_base_dir, dir );
}

/*===================================================================*/

	int
overlay_select_filename()
{
	int	message;
	char	*filename;

	message = file_select( &filename, overlay_base_dir );
	if( message == MESSAGE_OK ) {
		strcpy( overlay_filename, filename );
		options_set_overlay_filename( overlay_filename );
		if( options.overlay->doit )
			do_overlay();
		}

	return( message );
}

/******************************************************************************
 * Generate an overlay from data in an overlay file.
 */
	int *
gen_overlay( View *v, char *overlay_fname )
{
	FILE	*f;
	char	err_mess[1024], line[80], *id_string="NCVIEW-OVERLAY";
	float	x, y, version;
	long	i, j;
	size_t	x_size, y_size;
	int	*overlay;
	NCDim	*dim_x, *dim_y;

	/* Open the overlay file */
	if( (f = fopen(overlay_fname, "r")) == NULL ) {
		sprintf( err_mess, "Error: can't open overlay file named \"%s\"\n", 
			overlay_fname );
		in_error( err_mess );
		options_set_overlay_state( FALSE );
		return( NULL );
		}

	/* Make sure it is a valid overlay file
	 */
	if( fgets(line, 80, f) == NULL ) {
		sprintf( err_mess, "Error trying to read overlay file named \"%s\"\n",
			overlay_fname );
		in_error( err_mess );
		options_set_overlay_state( FALSE );
		return( NULL );
		}
	for( i=0; i<strlen(id_string); i++ )
		if( line[i] != id_string[i] ) {
			sprintf( err_mess, "Error trying to read overlay file named \"%s\"\nFile does not start with \"%s version-num\"\n", 
				overlay_fname, id_string );
			in_error( err_mess );
			options_set_overlay_state( FALSE );
			return( NULL );
			}
	sscanf( line, "%*s %f", &version );
	if( (version < 0.95) || (version > 1.05)) {
		sprintf( err_mess, "Error, overlay file has unknown version number: %f\nI am set up for version 1.0\n", version );
		in_error( err_mess );
		options_set_overlay_state( FALSE );
		return( NULL );
		}

	dim_x = *(v->variable->dim + v->x_axis_id);
	dim_y = *(v->variable->dim + v->y_axis_id);

	x_size = *(v->variable->size + v->x_axis_id);
	y_size = *(v->variable->size + v->y_axis_id);

	overlay = (int *)malloc( x_size*y_size*sizeof(int) );
	if( overlay == NULL ) {
		in_error( "Malloc of overlay field failed\n" );
		return( NULL );
		}
	for( i=0; i<x_size*y_size; i++ )
		*(overlay+i) = 0;

	/* Read in the overlay file -- skip lines with first char of #, 
	 * they are comments.
	 */
	while( fgets(line, 80, f) != NULL ) 
		if( line[0] != '#' ) {
			sscanf( line, "%f %f", &x, &y );
			i = gen_xform( x, x_size, dim_x->values );
			if( i == -2 ) 
				return( NULL );
			j = gen_xform( y, y_size, dim_y->values );
			if( j == -2 ) 
				return( NULL );
			if( (i > 0) && (j > 0)) 
				*(overlay + j*x_size + i) = 1;
			}

	return( overlay );
}

/******************************************************************************
 * Given the (dimensional) value from the overlay file, convert it to 
 * the nearest index along the proper dimension that the point corresponds to.
 */
	int
gen_xform( float value, int n, float *dimvals )
{
	float	min_dist, dist;
	int	i, min_place;

	min_dist  = 1.0e35;
	min_place = 0;

	/* See if off ends of dimvalues ... remember that it can be reversed */
	if( *dimvals > *(dimvals+n-1) ) {
		/* reversed */
		if( value > *dimvals )
			return( -1 );
		if( value < *(dimvals+n-1) )
			return( -1 );
		}
	else
		{
		if( value < *dimvals )
			return( -1 );
		if( value > *(dimvals+n-1) )
			return( -1 );
		}
	
	for( i=0; i<n; i++ ) {
		dist = fabs(*(dimvals+i) - value);
		if( dist < min_dist ) {
			min_dist  = dist;
			min_place = i;
			}
		}

	return( min_place );
}
