/*
	This provides an interface to the udunits package
	(http://www.unidata.ucar.edu/packages/udunits/index.html)
	that can be bypassed if desired.
*/

#include "ncview.includes.h"
#include "ncview.defines.h"
#include "math.h"

typedef struct {
	char	*name;
	void	*next;
} UniqList;

static int	valid_udunits_pkg;
static int	is_unique( char *units_name );
static UniqList *uniq = NULL;

/******************************************************************************/
#ifdef INC_UDUNITS

#include "udunits.h"

void udu_utinit( char *path )
{
	int	err;

	err = utInit( path );
	if( err == 0 )
		valid_udunits_pkg = 1;
	else
		{
		valid_udunits_pkg = 0;
		switch (err) {
			case UT_ENOFILE:
				fprintf( stderr, "Note: Udunits units file not found; no units conversion will be attmpted.\n" );
				fprintf( stderr, "(put path of units file in environmental variable UDUNITS_PATH)\n");
				break;

			case UT_ESYNTAX:
				fprintf( stderr, "Note: syntax error in udunits file; no units conversion will be attmpted.\n" );
				break;

			case UT_EUNKNOWN:
				fprintf( stderr, "Note: unknown specification in udunits file; no units conversion will be attmpted.\n" );
				break;

			case UT_EIO:
				fprintf( stderr, "Note: I/O error while reading udunits file; no units conversion will be attmpted.\n" );
				break;

			case UT_EALLOC:
				fprintf( stderr, "Note: Allocation error while reading udunits file; no units conversion will be attmpted.\n" );
				break;

			}
		}
}

int udu_utistime( char *dimname, char *units )
{
	utUnit	unit;

	if( units == NULL )
		return(0);

	if( valid_udunits_pkg ) {
		if( utScan( units, &unit ) != 0 ) {
			/* can't parse unit spec */
			if( is_unique(units) ) 
				fprintf( stderr, "Note: udunits: unknown units for %s: \"%s\"\n",
					dimname, units );
			return( 0 );
			}
		else
			{
			return( utHasOrigin( &unit) && utIsTime( &unit ));
			}
		}
	else
		return( 0 );
}

int udu_calc_tgran( int fileid, NCDim *d )
{
	utUnit	unit, seconds;
	double	slope, intercept;
	int	retval;

	if( ! valid_udunits_pkg )
		return( 0 );

	if( utScan( d->units, &unit ) != 0 ) {
		fprintf( stderr, "internal error: udu_calc_tgran with invalid unit string: %s\n",
			d->units );
		exit( -1 );
		}

	if( utScan( "seconds", &seconds ) != 0 ) {
		fprintf( stderr, "internal error: udu_calc_tgran can't parse seconds unit string!\n" );
		exit( -1 );
		}
	
	if( utConvert( &unit, &seconds, &slope, &intercept ) != 0 ) {
		fprintf( stderr, "internal error: udu_calc_tgran can't convert time units to seconds!\n" );
		exit( -1 );
		}
	
	if( fabs(slope - 1.) < 0.01 ) 
		retval = TGRAN_SEC;
	else if( fabs(slope - 60.) < .1 ) 
		retval = TGRAN_MIN;
	else if( fabs(slope - 3600.) < 1. ) 
		retval = TGRAN_HOUR;
	else if( fabs(slope - 86400.) < 10. ) 
		retval = TGRAN_DAY;
	else if( fabs(slope - 2629743.) < 1000. ) 
		retval = TGRAN_MONTH;
	else if( fabs(slope - 31556926.) < 10000. ) 
		retval = TGRAN_YEAR;
	else
		retval = TGRAN_SEC; /* don't know any better */

	return( retval );
}

void udu_fmt_time( char *temp_string, double new_dimval, NCDim *dim )
{
	utUnit	dataunits;
	int	year, month, day, hour, minute;
	float	second;
	static	char months[12][4] = { "Jan\0", "Feb\0", "Mar\0", "Apr\0",
				       "May\0", "Jun\0", "Jul\0", "Aug\0",
				       "Sep\0", "Oct\0", "Nov\0", "Dec\0"};

	if( ! valid_udunits_pkg ) {
		sprintf( temp_string, "%lg", new_dimval );
		return;
		}

	if( utScan( dim->units, &dataunits ) != 0 ) {
		fprintf( stderr, "internal error: udu_fmt_time can't parse data unit string!\n" );
		fprintf( stderr, "problematic units: \"%s\"\n", dim->units );
		fprintf( stderr, "dim->name: %s  dim->timelike: %d\n", dim->name, dim->timelike );
		exit( -1 );
		}

	if( utCalendar( new_dimval, &dataunits, &year, &month, &day, &hour, 
				&minute, &second ) != 0 ) {
		fprintf( stderr, "internal error: udu_fmt_time can't convert to calendar value!\n");
		fprintf( stderr, "units: >%s<\n", dim->units );
		exit( -1 );
		}
	
	if( dim->tgran == TGRAN_YEAR )
		sprintf( temp_string, "%04d", year );
	else if( dim->tgran == TGRAN_MONTH ) 
		sprintf( temp_string, "%s-%04d", months[month-1], year );
	else if( dim->tgran == TGRAN_DAY ) 
		sprintf( temp_string, "%1d-%s-%04d", day, months[month-1], year );
	else if( dim->tgran == TGRAN_HOUR )
		sprintf( temp_string, "%1d-%s-%04d %02d00", day, months[month-1], year, hour );
	else if( dim->tgran == TGRAN_MIN )
		sprintf( temp_string, "%1d-%s-%04d %02d:%02d", day, 
					months[month-1], year, hour, minute );
	else
		sprintf( temp_string, "%1d-%s-%04d %02d:%02d:%02.0f",
			day, months[month-1], year, hour, minute, second );
}

	static int 
is_unique( char *units )
{
	UniqList	*ul, *ul_new, *prev_ul;

	if( uniq == NULL ) {
		ul_new = (UniqList *)malloc( sizeof(UniqList) );
		ul_new->name = (char *)malloc(strlen(units)+1);
		strcpy( ul_new->name, units );
		ul_new->next = NULL;
		uniq = ul_new;
		return( TRUE );
		}
	
	ul = uniq;
	while( ul != NULL ) {
		if( strcmp( ul->name, units ) == 0 ) 
			return( FALSE );
		prev_ul = ul;
		ul = ul->next;
		}

	ul_new = (UniqList *)malloc( sizeof(UniqList) );
	ul_new->name = (char *)malloc(strlen(units)+1);
	strcpy( ul_new->name, units );
	ul_new->next = NULL;
	prev_ul->next = ul_new;
	return( TRUE );
}

/******************************************************************************/
#else

void udu_utinit( char *path )
{
	;
}

int udu_utistime( char *units )
{
	return( 0 );
}

int udu_calc_tgran( char *units )
{
	return( 0 );
}

void udu_fmt_time( char *temp_string, float new_dimval, NCDim *dim )
{
	sprintf( temp_string, "%g", new_dimval );
}

#endif
