/*
 * Ncview by David W. Pierce.  A visual netCDF file viewer.
 * Copyright (C) 1993 through 2003  David W. Pierce
 *
 * This program  is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 1 as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License, version 1, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * David W. Pierce
 * 6259 Caminito Carrean
 * San Diego, CA   92122
 * pierce@cirrus.ucsd.edu
 */

/*******************************************************************************
 * 	util.c
 *
 *	utility routines for ncview
 *
 *	should be independent of both the user interface and the data
 * 	file format.
 *******************************************************************************/

#include "ncview.includes.h"
#include "ncview.defines.h"
#include "ncview.protos.h"

#include "math.h"

extern Options   options;
extern NCVar     *variables;
extern ncv_pixel *pixel_transform;
extern FrameStore framestore;

static void get_min_max_onestep( NCVar *var, size_t n_other, size_t tstep, float *data, 
					float *min, float *max, int verbose );
static void handle_time_dim( int fileid, NCVar *v, NCDim *d );
static int  months_calc_tgran( int fileid, NCDim *d );


/* Variables local to routines in this file */
static  char    *month_name[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

/*******************************************************************************
 * Determine whether the data is "close enough" to the fill value
 */
int
close_enough( float data, float fill )
{
	float	criterion, diff;
	int	retval;

	if( fill == 0.0 )
		criterion = 1.0e-5;
	else if( fill < 0.0 )
		criterion = -1.0e-5*fill;
	else
		criterion = 1.0e-5*fill;

	diff = data - fill;
	if( diff < 0.0 ) 
		diff = -diff;

	if( diff <= criterion )
		retval = 1;
	else
		retval = 0;

 /* printf( "d=%f f=%f c=%f r=%d\n", data, fill, criterion, retval );  */
	return( retval );
}

/*******************************************************************************
 * Adds the given string to the list, and returns a pointer to the 
 * new list element.
 */
	Stringlist *
add_to_stringlist( Stringlist **list, char *new_string, void *aux )
{
	Stringlist 	*cursor, *new_el;
	int		i;

	new_stringlist( &new_el );
	new_el->string = (char *)malloc( strlen( new_string )+1);
	if( new_el == NULL ) {
		fprintf( stderr, "ncview: add_to_stringlist: malloc failed\n" );
		fprintf( stderr, "string trying to add: %s\n", new_string );
		exit( -1 );
		}
	strcpy( new_el->string, new_string );
	new_el->aux    = aux;

	i = 0;
	if( *list == NULL ) {
		*list = new_el;
		new_el->prev = NULL;
		}
	else
		{
		i = 1;
		cursor = *list;
		while( cursor->next != NULL ) {
			cursor = cursor->next;
			i++;
			}
		cursor->next = new_el;
		new_el->prev = cursor;
		}
	new_el->index = i;

	return( new_el );
}

/*******************************************************************************
 * Adds the given string to the list, and returns a pointer to the 
 * new list element, with alphabetic ordering.
 */
	Stringlist *
add_to_stringlist_ordered( Stringlist **list, char *new_string, void *aux )
{
	Stringlist 	*cursor, *new_el, *prev_el;
	int		i;

	new_stringlist( &new_el );
	new_el->string = (char *)malloc( strlen( new_string )+1);
	if( new_el == NULL ) {
		fprintf( stderr, "ncview: add_to_stringlist: malloc failed\n" );
		fprintf( stderr, "string trying to add: %s\n", new_string );
		exit( -1 );
		}
	strcpy( new_el->string, new_string );
	new_el->aux    = aux;

	i = 0;
	if( *list == NULL ) {
		*list = new_el;
		new_el->prev = NULL;
		}
	else
		{
		i = 1;
		cursor = *list;
		prev_el = NULL;
		while( (cursor != NULL) && (strcmp( new_string, cursor->string) > 0)) {
			prev_el = cursor;
			cursor  = cursor->next;
			i++;
			}
		if( cursor == NULL ) {
			prev_el->next = new_el;
			new_el->prev  = prev_el;
			}
		else if( prev_el == NULL ) {
			*list = new_el;
			new_el->next = cursor;
			cursor->prev = new_el;
			}
		else
			{
			new_el->next = cursor;
			cursor->prev = new_el;
			prev_el->next = new_el;
			new_el->prev  = prev_el;
			}
		}
	new_el->index = i;

	return( new_el );
}


/*******************************************************************************
 * Concatenate onto a stringlist
 */
	void
sl_cat( Stringlist **dest, Stringlist **src )
{
	Stringlist *sl;

	sl = *src;
	while( sl != NULL ) {
		add_to_stringlist( dest, sl->string, NULL );
		sl = sl->next;
		}
}

/*******************************************************************************
 * Allocate space for a new Stringlist element 
 */
	void
new_stringlist( Stringlist **el )
{
	(*el)       = (Stringlist *)malloc( sizeof( Stringlist ));
	(*el)->next = NULL;
}

/******************************************************************************
 * Add the passed NCVar element to the list 
 */
	void
add_to_varlist( NCVar **list, NCVar *new_el )
{
	int	i;
	NCVar	*cursor;

	i = 0;
	if( *list == NULL ) {
		*list = new_el;
		new_el->prev = NULL;
		}
	else
		{
		i = 1;
		cursor = *list;
		while( cursor->next != NULL ) {
			cursor = cursor->next;
			i++;
			}
		cursor->next = new_el;
		new_el->prev = cursor;
		}
}

/******************************************************************************
 * Allocate space for a new NCVar element
 */
	void
new_variable( NCVar **el )
{
	(*el)       = (NCVar *)malloc( sizeof( NCVar ));
	(*el)->next = NULL;
}


/******************************************************************************
 * Allocate space for a new FDBlist element
 */
	void
new_fdblist( FDBlist **el )
{
	NetCDFOptions	*new_netcdf_options;

	(*el)           = (FDBlist *)malloc( sizeof( FDBlist ));
	(*el)->next     = NULL;
	(*el)->filename = (char *)malloc( 1024 );
	strcpy( (*el)->filename, "UNINITIALIZED" );

	new_netcdf( &new_netcdf_options );
	(*el)->aux_data = new_netcdf_options;
}

/******************************************************************************
 * Allocate space for a NetCDFOptions structure.
 */
	void
new_netcdf( NetCDFOptions **n )
{
	(*n) = (NetCDFOptions *)malloc( sizeof( NetCDFOptions ));
	(*n)->valid_range_set  = FALSE;
	(*n)->valid_min_set    = FALSE;
	(*n)->valid_max_set    = FALSE;
	(*n)->scale_factor_set = FALSE;
	(*n)->add_offset_set   = FALSE;

	(*n)->valid_range[0] = 0.0;
	(*n)->valid_range[1] = 0.0;
	(*n)->valid_min      = 0.0;
	(*n)->valid_max      = 0.0;
	(*n)->scale_factor   = 1.0;
	(*n)->add_offset     = 0.0;
}

/******************************************************************************
 * What's in this stringlist, anyway?
 */
	void
dump_stringlist( Stringlist *s )
{
	while( s != NULL ) {
		printf( "%d: %s\n", s->index, s->string );
		s = s->next;
		}
}

/******************************************************************************
 * Scale the data, replicate it, and convert to a pixel type array.  I'm afraid
 * that for speed, this considers 'ncv_pixel' to be a single byte value.  Make sure
 * to change it if you change the definition of ncv_pixel!  Returns 0 on
 * success, -1 on failure.
 */
	int
data_to_pixels( View *v )
{
	long	i, j, j2;
	size_t	x_size, y_size, big_x_size, big_y_size;
	ncv_pixel pix_val;
	float	data_range, rawdata, data, fill_value, *big_data;
	long	blowup, result, orig_minmax_method;
	char	error_message[1024];

	/* Make sure the limits have been set on this variable.
	 * They won't always be because an initial expose event can 
	 * cause this routine to be executed before the min and
	 * maxes are calcuclated.
	 */
	if( ! v->variable->have_set_range )
		return( -1 );

	blowup   = options.blowup;

	x_size     = *(v->variable->size + v->x_axis_id);
	y_size     = *(v->variable->size + v->y_axis_id);
	big_x_size = x_size*blowup;
	big_y_size = y_size*blowup;
	big_data   = (float *)malloc( big_x_size*big_y_size*sizeof(float));
	if( big_data == NULL ) {
		fprintf( stderr, "ncview: data_to_pixels: can't allocate data expansion array\n" );
		fprintf( stderr, "requested size: %ld bytes\n", big_x_size*big_y_size*sizeof(float) );
		fprintf( stderr, "big_x_size, big_y_size, float_size: %ld %ld %ld\n", 
				big_x_size, big_y_size, sizeof(float) );
		fprintf( stderr, "blowup: %d\n", options.blowup );
		exit( -1 );
		}

	/* If we are doing overlays, implement them */
	if( options.overlay->doit && (options.overlay->overlay != NULL)) {
		for( i=0; i<(x_size*y_size); i++ ) {
			*((float *)v->data + i) = 
			     (float)(1 - *(options.overlay->overlay+i)) * *((float *)v->data + i) +
			     (float)(*(options.overlay->overlay+i)) * v->variable->fill_value;
			}
		}

	expand_data( big_data, v );
	data_range = v->variable->user_max - v->variable->user_min;
	fill_value = v->variable->fill_value;

	if( (v->variable->user_max == 0) &&
	    (v->variable->user_min == 0) ) {
		in_set_cursor_normal();
		in_button_pressed( BUTTON_PAUSE, MOD_1 );
		if( options.min_max_method == MIN_MAX_METHOD_EXHAUST ) {
	    		sprintf( error_message, "min and max both 0 for variable %s.\n(Checked all data)", 
								v->variable->name );
			in_error( error_message );
			return( -1 );
			}
	    	sprintf( error_message, "min and max both 0 for variable %s.\nI can check ALL the data instead of subsampling if that's OK,\nor just cancel viewing this variable.",
	    				v->variable->name );
		result = in_dialog( error_message, NULL, TRUE );
		if( result == MESSAGE_OK ) {
			orig_minmax_method = options.min_max_method;
			options.min_max_method = MIN_MAX_METHOD_EXHAUST;
			init_min_max( v->variable );
			options.min_max_method = orig_minmax_method;
			if( (v->variable->user_max == 0) &&
	    		    (v->variable->user_min == 0) ) {
	    			sprintf( error_message, "min and max both 0 for variable %s.\n(Checked all data)", 
								v->variable->name );
				in_error( error_message );
				return( -1 );
				}
			else
				return( data_to_pixels(v) );
			}
		else
			return( -1 );
	    	}

	if( v->variable->user_max == v->variable->user_min ) {
		in_set_cursor_normal();
	    	sprintf( error_message, "min and max both %g for variable %s",
	    		v->variable->user_min, v->variable->name );
		x_error( error_message );
		return( -1 );
	    	}

	for( j=0; j<big_y_size; j++ ) {

		if( options.invert_physical )
			j2 = j;
		else
			j2 = big_y_size - j - 1;

		for( i=0; i<big_x_size; i++ ) {
			rawdata =  *(big_data + i + j2*big_x_size);
			if( close_enough(rawdata, fill_value) || (rawdata == FILL_FLOAT))
				pix_val = *pixel_transform;
			else
				{
				data = (rawdata - v->variable->user_min) / data_range;
				clip_f( &data, 0.0, .9999 );
				switch( options.transform ) {
					case TRANSFORM_NONE:	break;

					/* This might cause problems.  It is at odds with what
					 * the manual claims--at least for Ultrix--but works, 
					 * whereas what the manual claims works, doesn't!
					 */
					case TRANSFORM_LOW:	data = sqrt( data );  
								data = sqrt( data );
								break;

					case TRANSFORM_HI:	data = data*data*data*data;     break;
					}		
				pix_val = (ncv_pixel)(data * options.n_colors) + 10;
				if( options.invert_colors )
					pix_val = options.n_colors - pix_val + 10;
				if( options.display_type == PseudoColor )
					pix_val = *(pixel_transform+pix_val);
				}
			*(v->pixels + i + j*big_x_size) = pix_val;
			}
		}
	free( big_data );
	return( 0 );
}

/******************************************************************************
 * Returns the number of entries in the NCVarlist
 */
	int
n_vars_in_list( NCVar *v )
{
	NCVar *cursor;
	int   i;

	i      = 0;
	cursor = v;
	while( cursor != NULL ) {
		i++;
		cursor = cursor->next;
		}

	return( i );
}

/******************************************************************************
 * Returns the number of entries in the Stringlist
 */
	int
n_strings_in_list( Stringlist *s )
{
	Stringlist *c;
	int	   i;

	i = 0;
	c = s;
	while( c != NULL ) {
		i++;
		c = c->next;
		}

	return( i );
}
	
/******************************************************************************
 * Given a list of variable *names*, initialize the variable *structure* and
 * add it to the global list of variables.
 */
	void
add_vars_to_list( Stringlist *var_list, int id, char *filename )
{
	Stringlist *var;

	if( options.debug )
		fprintf( stderr, "adding vars to list for file %s\n", filename );
	var = var_list;
	while( var != NULL ) {
		if( options.debug ) 
			fprintf( stderr, "adding variable %s to list\n", var->string );
		add_var_to_list( var->string, id, filename );
		var = var->next;
		}
	if( options.debug ) 
		fprintf( stderr, "done adding vars for file %s\n", filename );
}

/******************************************************************************
 * For the given variable name, fill out the variable and file structures,
 * and add them into the global variable list.
 */
	void
add_var_to_list( char *var_name, int file_id, char *filename )
{
	NCVar	*var, *new_var;
	int	n_dims;
	FDBlist	*new_fdb, *fdb;

	/* make a new file description entry for this var/file combo */
	new_fdblist( &new_fdb );
	new_fdb->id       = file_id;
	new_fdb->var_size = fi_var_size( file_id, var_name );
	strcpy( new_fdb->filename, filename );

	/* fill out auxilliary (data-file format dependent) information
	 * for the new fdb.
	 */
	fi_fill_aux_data( file_id, var_name, new_fdb );

	/* Does this variable already have an entry on the var list? */
	var = get_var( var_name );
	if( var == NULL ) {	/* NO -- make a new NCVar structure */
		new_variable( &new_var );
		new_var->name       = (char *)malloc( strlen(var_name)+1 );
		strcpy( new_var->name, var_name );
		n_dims              = fi_n_dims( file_id, var_name );
		new_var->n_dims     = n_dims;
		if( options.debug )
			fprintf( stderr, "adding variable %s with %d dimensions\n",
				var_name, n_dims );
		new_var->first_file = new_fdb;
		new_var->last_file  = new_fdb;
		new_var->global_min = 0.0;
		new_var->global_max = 0.0;
		new_var->user_min   = 0.0;
		new_var->user_max   = 0.0;
		new_var->have_set_range = FALSE;
		new_var->size       = fi_var_size( file_id, var_name );
		new_var->fill_value = DEFAULT_FILL_VALUE;
		fi_fill_value( new_var, &(new_var->fill_value) );
		new_fdb->prev       = NULL;
		fill_dim_structs( new_var );
		add_to_varlist  ( &variables, new_var );
		new_var->is_virtual = FALSE;
		}
	else	/* YES -- just add the FDB to the list of files in which 
		 * this variable appears, and accumulate the variable's size.
		 */
		{
		/* Go to the end of the file list and add it there */
		if( options.debug )
			fprintf( stderr, "adding another file with variable %s in it\n",
				var_name );
		if( var->last_file == NULL ) {
			fprintf( stderr, "ncview: add_var_to_list: internal ");
			fprintf( stderr, "inconsistancy; var has no last_file\n" );
			exit( -1 );
			}
		fdb = var->first_file;
		while( fdb->next != NULL )
			fdb = fdb->next;
		fdb->next         = new_fdb;
		new_fdb->prev     = fdb;
		var->last_file    = new_fdb;
		*(var->size)      += *(new_fdb->var_size);
		var->is_virtual   = TRUE;
		}
}

/******************************************************************************
 * Calculate the min and max values for the passed variable.
 */
	void
init_min_max( NCVar *var )
{
	long	n_other, i, step;
	size_t	n_timesteps;
	float	*data, init_min, init_max;
	int	verbose;

	init_min =  9.9e30;
	init_max = -9.9e30;
	var->global_min = init_min;
	var->global_max = init_max;

	printf( "calculating min and maxes for %s", var->name );

	/* n_other is the number of elements in a single timeslice of the data array */
	n_timesteps = *(var->size);
	n_other     = 1L;
	for( i=1; i<var->n_dims; i++ )
		n_other *= *(var->size+i);

	data = (float *)malloc( n_other * sizeof( float ));
	if( data == NULL ) {
		fprintf( stderr, "ncview: init_min_max_file: failed on malloc of data array\n" );
		exit( -1 );
		}

	/* We always get the min and max of the first, middle, and last time 
	 * entries if they are distinct.
	 */
	verbose = TRUE;
	step    = 0L;
	get_min_max_onestep( var, n_other, step, data, 
			&(var->global_min), &(var->global_max), verbose );
	if( n_timesteps == 1 ) {
		if( verbose )
			printf( "\n" );
		free( data );
		check_ranges( var );
		return;
		}

	step = n_timesteps-1L;
	get_min_max_onestep( var, n_other, step, data, 
			&(var->global_min), &(var->global_max), verbose );
	if( n_timesteps == 2 ) {
		if( verbose )
			printf( "\n" );
		free( data );
		check_ranges( var );
		return;
		}

	step = (n_timesteps-1L)/2L;
	get_min_max_onestep( var, n_other, step, data, 
			&(var->global_min), &(var->global_max), verbose );
	if( n_timesteps == 3 ) {
		if( verbose )
			printf( "\n" );
		free( data );
		check_ranges( var );
		return;
		}

	switch( options.min_max_method ) {
		case MIN_MAX_METHOD_FAST: 
			if( verbose )
				printf( "\n" );
			break;
			
		case MIN_MAX_METHOD_MED:     
			verbose = TRUE;
			step = (n_timesteps-1L)/4L;
			get_min_max_onestep( var, n_other, step, data, 
				&(var->global_min), &(var->global_max), verbose );
			step = (3L*(n_timesteps-1L))/4L;
			get_min_max_onestep( var, n_other, step, data, 
				&(var->global_min), &(var->global_max), verbose );
			if( verbose )
				printf( "\n" );
			break;
				
		case MIN_MAX_METHOD_SLOW:
			verbose = TRUE;
			for( i=2; i<=9; i++ ) { 
				printf( "." );
				step = (i*(n_timesteps-1L))/10L;
				get_min_max_onestep( var, n_other, step, data, 
					&(var->global_min), &(var->global_max), verbose );
				}
			if( verbose )
				printf( "\n" );
			break;
			
		case MIN_MAX_METHOD_EXHAUST:
			verbose = TRUE;
			for( i=1; i<(n_timesteps-2L); i++ ) {
				step = i;
				get_min_max_onestep( var, n_other, step, data, 
					&(var->global_min), &(var->global_max), verbose );
				}
			if( verbose )
				printf( "\n" );
			break;
		}

	if( (var->global_min == init_min) && (var->global_max == init_max) ) {
		var->global_min = 0.0;
		var->global_max = 0.0;
		}
		
	check_ranges( var );
	free( data );
}

/******************************************************************************
 * Try to reconcile the computed and specified (if any) data range
 */
	void
check_ranges( NCVar *var )
{
	float	min, max;
	int	message;
	char	temp_string[ 1024 ];

	if( netcdf_min_max_option_set( var, &min, &max ) ) {
		if( var->global_min < min ) {
			sprintf( temp_string, "Calculated minimum (%g) is less than\nvalid_range minimum (%g).  Reset\nminimum to valid_range minimum?", var->global_min, min );
			message = in_dialog( temp_string, NULL, TRUE );
			if( message == MESSAGE_OK )
				var->global_min = min;
			}
		if( var->global_max > max ) {
			sprintf( temp_string, "Calculated maximum (%g) is greater\nthan valid_range maximum (%g). Reset\nmaximum to valid_range maximum?", var->global_max, max );
			message = in_dialog( temp_string, NULL, TRUE );
			if( message == MESSAGE_OK )
				var->global_max = max;
			}
		}

	if( netcdf_min_option_set( var, &min ) ) {
		if( var->global_min < min ) {
			sprintf( temp_string, "Calculated minimum (%g) is less than\nvalid_min minimum (%g).  Reset\nminimum to valid_min value?", var->global_min, min );
			message = in_dialog( temp_string, NULL, TRUE );
			if( message == MESSAGE_OK )
				var->global_min = min;
			}
		}

	if( netcdf_max_option_set( var, &max ) ) {
		if( var->global_max > max ) {
			sprintf( temp_string, "Calculated maximum (%g) is greater than\nvalid_max maximum (%g).  Reset\nmaximum to valid_max value?", var->global_max, max );
			message = in_dialog( temp_string, NULL, TRUE );
			if( message == MESSAGE_OK )
				var->global_max = max;
			}
		}

	var->user_min = var->global_min;
	var->user_max = var->global_max;
	var->have_set_range = TRUE;
}

/******************************************************************************
 * get_min_max utility routine; is passed timestep number where want to 
 * determine extrema 
 */
	static void
get_min_max_onestep( NCVar *var, size_t n_other, size_t tstep, float *data, 
					float *min, float *max, int verbose )
{
	size_t	*start, *count, n_time;
	size_t	j;
	int	i;
	float	dat, fill_v;
	
	count  = (size_t *)malloc( var->n_dims * sizeof( size_t ));
	start  = (size_t *)malloc( var->n_dims * sizeof( size_t ));
	fill_v = var->fill_value;

	n_time = *(var->size);
	if( tstep > (n_time-1) )
		tstep = n_time-1;

	*(count) = 1L;
	*(start) = tstep;
	for( i=1; i<var->n_dims; i++ ) {
		*(start+i) = 0L;
		*(count+i) = *(var->size + i);
		}

	if( verbose ) {
		printf( "." );
		fflush( stdout );
		}

	fi_get_data( var, start, count, data );

	for( j=0; j<n_other; j++ ) {
		dat = *(data+j);
		if( dat != dat )
			dat = fill_v;
		if( (! close_enough(dat, fill_v)) && (dat != FILL_FLOAT) )
			{
			if( dat > *max )
				*max = dat;
			if( dat < *min )
				*min = dat;
			}
		}
		
	free( count );
	free( start );
}

/******************************************************************************
 * convert a variable name to a NCVar structure
 */
	NCVar *
get_var( char *var_name )
{
	NCVar	*ret_val;

	ret_val = variables;
	while( ret_val != NULL )
		if( strcmp( var_name, ret_val->name ) == 0 )
			return( ret_val );
		else
			ret_val = ret_val->next;

	return( NULL );
}

/******************************************************************************
 * Clip out of range floats 
 */
	void
clip_f( float *data, float min, float max )
{
	if( *data < min )
		*data = min;
	if( *data > max )
		*data = max;
}

/******************************************************************************
 * Turn a virtual variable 'place' array into a file/place pair.  Which is
 * to say, the virtual size of a variable spans the entries in all the files; 
 * the actual place where the entry for a particular virtual location can
 * be found is in a file/actual_place pair.  This routine does the conversion.
 * Note that this routine is assuming the netCDF convention that ONLY THE
 * FIRST index can be contiguous across files.  The first index is typically
 * the time index in netCDF files.  NOTE! that 'act_pl' must be allocated 
 * before calling this!
 */
	void
virt_to_actual_place( NCVar *var, size_t *virt_pl, size_t *act_pl, FDBlist **file )
{
	FDBlist	*f;
	size_t	v_place, cur_start, cur_end;
	int	i, n_dims;

	f       = var->first_file;
	n_dims  = fi_n_dims( f->id, var->name );
	v_place = *(virt_pl);

	if( v_place >= *(var->size) ) {
		fprintf( stderr, "ncview: virt_to_actual_place: error trying ");
		fprintf( stderr, "to convert the following virtual place to\n" );
		fprintf( stderr, "an actual place for variable %s:\n", var->name );
		for( i=0; i<n_dims; i++ )
			fprintf( stderr, "[%1d]: %ld\n", i, *(virt_pl+i) );
		exit( -1 );
		}

	cur_start = 0L;
	cur_end   = *(f->var_size) - 1L;
	while( v_place > cur_end ) {
		cur_start += *(f->var_size);
		f          = f->next;
		cur_end   += *(f->var_size);
		}

	*file = f;
	*act_pl = v_place - cur_start;

	/* Copy the rest of the indices over */
	for( i=1; i<n_dims; i++ )
		*(act_pl+i) = *(virt_pl+i);
}

/***************************************************************************
 * Change the current data transformation
 */
	void
change_transform( int delta )
{
	options.transform += delta;
	if( options.transform > N_TRANSFORMS )
		options.transform = 1;
	if( options.transform < 1 )
		options.transform = N_TRANSFORMS;

	switch( options.transform ) {
		case TRANSFORM_NONE: in_set_label( LABEL_TRANSFORM, "Linear" ); break;
		case TRANSFORM_LOW : in_set_label( LABEL_TRANSFORM, "Low"    ); break;
		case TRANSFORM_HI  : in_set_label( LABEL_TRANSFORM, "Hi"     ); break;
		default:
			fprintf( stderr, "ncview: change_transform: unknown transform %d\n",
				options.transform );
			exit( -1 );
		}

	view_draw( TRUE );
}

/******************************************************************************
 * Initialize all the fields in the dim structure by reading from the data file
 */
	void
fill_dim_structs( NCVar *v )
{
	int	i, fileid;
	NCDim	*d;
	char	*dim_name;

	fileid = v->first_file->id;
	v->dim = (NCDim **)malloc( v->n_dims*sizeof( NCDim *));
	for( i=0; i<v->n_dims; i++ ) {
		if( is_scannable( v, i ) ) {
			dim_name     = fi_dim_id_to_name( fileid, v->name, i );
			*(v->dim + i)= (NCDim *)malloc( sizeof( NCDim ));
			d            = *(v->dim+i);
			d->name      = dim_name;
			d->long_name = fi_dim_longname( fileid, dim_name );
			d->units     = fi_dim_units   ( fileid, dim_name );
			d->size      = *(v->size+i);
			handle_time_dim( fileid, v, d );
			if( options.debug ) 
				fprintf( stderr, "adding scannable dim to var %s: dimname: %s dimsize: %ld\n", v->name, dim_name, d->size );
			}
		else
			{
			/* Indicate non-scannable dimensions by a NULL */
			*(v->dim + i) = NULL;
			if( options.debug ) 
				fprintf( stderr, "adding non-scannable dim to var %s\n", v->name );
			}
		}
}

/******************************************************************************
 * Is this variable a "scannable" variable -- i.e., accessable by the taperecorder
 * style buttons?
 */
	int
is_scannable( NCVar *v, int i )
{
	/* The unlimited record dimension is always scannable */
	if( i == 0 )
		return( TRUE );

	if( *(v->size+i) > 1 )
		return( TRUE );
	else
		return( FALSE );
}

/******************************************************************************
 * Calculate the minimum and maximum values in the dimension structs 
 */
	void
calc_dim_minmaxes()
{
	int	i, j;
	NCVar	*v;
	NCDim	*d;
	char	temp_str[1024];
	nc_type	type;
	double	temp_double;

	v = variables;
	while( v != NULL ) {
		for( i=0; i<v->n_dims; i++ ) {
			d = *(v->dim+i);
			if( d != NULL ) {
				d->values = (float *)malloc(*(v->size+i)*sizeof(float));
				type = fi_dim_value( v, i, 0L, &temp_double, temp_str );
				if( type == NC_DOUBLE ) {
					d->values = (float *)malloc(*(v->size+i)*sizeof(float));
					for( j=0; j<*(v->size+i); j++ ) {
						type = fi_dim_value( v, i, j, &temp_double, temp_str );
						*(d->values+j) = (float)temp_double;
						}
					d->min  = *(d->values);
					d->max  = *(d->values + *(v->size+i) - 1);
					}
				else
					{
					if( options.debug ) 
						fprintf( stderr, "**Note: non-float dim found; i=%d\n", i );
					d->min  = 1.0;
					d->max  = *(v->size+i);
					for( j=0; j<*(v->size+i); j++ )
						*(d->values+j) = (float)j;
					}
				}
			}
		v = v->next;
		}
}
	
/******************************************************************************
 * Actually do the "blowup" of the FLOATING POINT (not pixel) data, converting 
 * it to the large version by either interpolation or replication.
 */
	void
expand_data( float *big_data, View *v )
{
	size_t	to_width, x_size, y_size;
	long	line, i, j, i2, j2;
	int	blowup, blowupsq;
	float	step;
	float	base_val, right_val, below_val, val, bupr;
	float	base_x, base_y, del_x, del_y;
	float	est1, est2, frac_x, frac_y;
	float 	fill_val;

	fill_val = v->variable->fill_value;
	x_size   = *(v->variable->size + v->x_axis_id);
	y_size   = *(v->variable->size + v->y_axis_id);
	
	blowup   = options.blowup;
	blowupsq = blowup * blowup;
	to_width = blowup * x_size;
	if( (blowupsq < blowup) || (to_width < blowup) ) {
		fprintf( stderr, "ncview: data_to_pixels: too much magnification\n" );
		fprintf( stderr, "blowup=%d  blowupsq=%d  to_width=%d\n",
				blowup, blowupsq, to_width );
		exit( -1 );
		}

	if( options.blowup_type == BLOWUP_REPLICATE ) { 
		for( j=0; j<y_size; j++ ) {
			for( i=0; i<x_size; i++ )
				for( i2=0; i2<blowup; i2++ )
					*(big_data + i*blowup + j*to_width*blowup + i2) = *((float *)((float *)v->data)+i+j*x_size);
			for( line=1; line<blowup; line++ )
				for( i2=0; i2<to_width; i2++ )
					*(big_data + i2 + j*to_width*blowup + line*to_width) =
						*(big_data + i2 + j*to_width*blowup);
			}
		} 
	else 	{ /* BLOWUP_BILINEAR */
		bupr = 1.0/(float)blowup;

		/* Horizontal base lines */
		for( j=0; j<y_size; j++ )
		for( i=0; i<x_size-1; i++ ) {
			base_val  = *((float *)v->data + i   + j*x_size);
			right_val = *((float *)v->data + i+1 + j*x_size);

			if( close_enough(base_val,  fill_val) || 
			    close_enough(right_val, fill_val))
				step = 0.0;
			else
				step = (right_val-base_val)*bupr;
			val = base_val;

			for( i2=0; i2 < blowup; i2++ ) {
				*(big_data + i*blowup+i2 + j*x_size*blowupsq ) = val;
				val += step;
				}
			}

		/* Vertical base lines */
		for( j=0; j<y_size-1; j++ )
		for( i=0; i<x_size; i++ ) {
			base_val  = *((float *)v->data + i   + j*x_size);
			below_val = *((float *)v->data + i + (j+1)*x_size);
			if( close_enough(base_val,  fill_val) || 
			    close_enough(below_val, fill_val))
				step = 0.0;
			else
				step = (below_val-base_val)*bupr;
			val = base_val;

			for( j2=0; j2 < blowup; j2++ ) {
				*(big_data + i*blowup + j*to_width*blowup + j2*to_width ) = val;
				val += step;
				}
			}
		/* Now, fill in the interior of the square by 
		 * interpolating from the horizontal and vertical
		 * base lines.
		 */
		for( j=0; j<y_size-1; j++ )
		for( i=0; i<x_size-1; i++ ) {
			for( j2=1; j2<blowup; j2++ )
			for( i2=1; i2<blowup; i2++ ) {
				frac_x = (float)i2*bupr;
				frac_y = (float)j2*bupr;
				base_x = *(big_data + i*blowup + j*to_width*blowup+j2*to_width);
				right_val = *(big_data + (i+1)*blowup + j*to_width*blowup+ j2*to_width);
				base_y = *(big_data + i*blowup+i2 + j*to_width*blowup);
				below_val = *(big_data + i*blowup+i2 + (j+1)*to_width*blowup);

				if( close_enough(base_x,    fill_val) || 
				    close_enough(right_val, fill_val) || 
				    (i == x_size-1) )
					del_x = 0.0;
				else
					del_x  = right_val - base_x;
				if( close_enough(base_y,    fill_val) || 
				    close_enough(below_val, fill_val) || 
				    (j == y_size-1) )
					del_y = 0.0;
				else
					del_y  = below_val - base_y;
				est1 = frac_x*del_x + base_x;
				est2 = frac_y*del_y + base_y;
				*(big_data + i*blowup+i2 + j*to_width*blowup + j2*to_width ) =
					(est1 + est2)*.5;
				}
			}
		/* Fill in right hand side */
		for( j=0; j<blowup*y_size; j++ )
		for( i=0; i<blowup; i++ )
			*(big_data + (x_size-1)*blowup+i + j*to_width ) = fill_val;

		/* Fill in bottom */
		for( j=0; j<blowup; j++ )
		for( i=0; i<blowup*x_size; i++ )
			*(big_data + i + (y_size-1)*to_width*blowup + j*to_width ) = fill_val;
		}
}

/******************************************************************************
 * Set the style of blowup we want to do.
 */
	void
set_blowup_type( int new_type )
{
	if( new_type == BLOWUP_REPLICATE ) 
		in_set_label( LABEL_BLOWUP_TYPE, "Repl"   );
	else
		in_set_label( LABEL_BLOWUP_TYPE, "Bi-lin" );

	options.blowup_type = new_type;
}

/******************************************************************************
 * If we allowed strings of arbitrary length, some of the widgets
 * would crash when trying to display them.
 */
	char *
limit_string( char *s )
{
	int	i;

	i = strlen(s)-1;
	while( *(s+i) == ' ' )
		i--;
	*(s+i+1) = '\0';

	if( strlen(s) > MAX_DISPLAYED_STRING_LENGTH )
		*(s+MAX_DISPLAYED_STRING_LENGTH) = '\0';

	return( s );
}

/******************************************************************************
 * If we try to print to an already existing file, then warn the user
 * before clobbering it.
 */
	int
warn_if_file_exits( char *fname )
{
	int	retval;
	FILE	*f;
	char	message[1024];

	if( (f = fopen(fname, "r")) == NULL )
		return( MESSAGE_OK );
	fclose(f);

	sprintf( message, "OK to overwrite existing file %s?\n", fname );
	retval = in_dialog( message, NULL, TRUE );
	return( retval );
}

/******************************************************************************/

	static void
handle_time_dim( int fileid, NCVar *v, NCDim *d )
{
	if( udu_utistime( d->name, d->units ) ) {
		d->timelike = 1;
		d->time_std = TSTD_UDUNITS;
		d->tgran    = udu_calc_tgran( fileid, d );
		}
	else if( epic_istime0( fileid, v, d )) {
		d->timelike = 1;
		d->time_std = TSTD_EPIC_0;
		d->tgran    = epic_calc_tgran( fileid, d );
		}
	else if( (d->units != NULL) &&
		 (strlen(d->units) >= 5) &&
		 (strncasecmp( d->units, "month", 5 ) == 0 ))  {
		d->timelike = 1;
		d->time_std = TSTD_MONTHS;
		d->tgran    = months_calc_tgran( fileid, d );
		}
	else
		d->timelike = 0;
}

/******************************************************************************/
	static int
months_calc_tgran( int fileid, NCDim *d )
{
	char	temp_string[128];
	float	delta, v0, v1;
	int	type;
	double	temp_double;

	if( d->size < 2 ) {
		return( TGRAN_DAY );
		}

	type = netcdf_dim_value( fileid, d->name, 0L, &temp_double, temp_string, 0L );
	if( type == NC_DOUBLE )
		v0 = (float)temp_double;
	else
		{
		fprintf( stderr, "Note: can't calculate time granularity, unrecognized timevar type (%d)\n",
			type );
		return( TGRAN_DAY );
		}

	type = netcdf_dim_value( fileid, d->name, 1L, &temp_double, temp_string, 1L );
	if( type == NC_DOUBLE )
		v1 = (float)temp_double;
	else
		{
		fprintf( stderr, "Note: can't calculate time granularity, unrecognized timevar type (%d)\n",
			type );
		return( TGRAN_DAY );
		}
	
	delta = v1 - v0;

	if( delta > 11.5 ) 
		return( TGRAN_YEAR );
	if( delta > .95 ) 
		return( TGRAN_MONTH );
	if( delta > .03 ) 
		return( TGRAN_DAY );

	return( TGRAN_MIN );
}

/******************************************************************************/
void fmt_time( char *temp_string, double new_dimval, NCDim *dim )
{
	int 	year, month, day;

	if( ! dim->timelike ) {
		fprintf( stderr, "ncview: internal error: fmt_time called on non-timelike axis!\n");
		fprintf( stderr, "dim name: %s\n", dim->name );
		exit( -1 );
		}

	if( dim->time_std == TSTD_UDUNITS ) 
		udu_fmt_time( temp_string, new_dimval, dim );

	else if( dim->time_std == TSTD_EPIC_0 ) 
		epic_fmt_time( temp_string, new_dimval, dim );

	else if( dim->time_std == TSTD_MONTHS ) {
		/* Format for months standard */
		year  = (int)( (new_dimval-1.0) / 12.0 );
		month = (int)( (new_dimval-1.0) - year*12 + .01 );
		month = (month < 0) ? 0 : month;
		month = (month > 11) ? 11 : month;
		day   =
		   (int)( ((new_dimval-1.0) - year*12 - month) * 30.0) + 1;
		sprintf( temp_string, "%s %2d %4d", month_name[month],
				day, year+1 );
		}
}

